#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/user.h>
#include <linux/malloc.h>
#include <linux/binfmts.h>
#include <linux/personality.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>

#define COM_MAGIC	0x4C434330	/* LCC0 */
#define COM_START	0x08000000
#define COM_DATA	COM_START
#define COM_BSS		COM_START

static int load_flat_binary(struct linux_binprm *, struct pt_regs * regs);

static struct linux_binfmt flat_format = {
	NULL,
	&__this_module,
	load_flat_binary,
	NULL,
	NULL
};

static unsigned long * create_flat_tables(char * p, struct linux_binprm * bprm)
{
	char **argv, **envp;
	unsigned long * sp;
	int argc = bprm->argc;
	int envc = bprm->envc;

	sp = (unsigned long *) ((-(unsigned long)sizeof(char *)) & (unsigned long) p);

	sp -= envc+1;
	envp = (char **) sp;
	sp -= argc+1;
	argv = (char **) sp;
	put_user((unsigned long) envp,--sp);
	put_user((unsigned long) argv,--sp);
	put_user(argc,--sp);
	current->mm->arg_start = (unsigned long) p;
	while (argc-->0) {
		char c;
		put_user(p,argv++);
		do {
			get_user(c,p++);
		} while (c);
	}
	put_user(NULL,argv);
	current->mm->arg_end = current->mm->env_start = (unsigned long) p;
	while (envc-->0) {
		char c;
		put_user(p,envp++);
		do {
			get_user(c,p++);
		} while (c);
	}
	put_user(NULL,envp);
	current->mm->env_end = (unsigned long) p;
	return sp;
}

static inline int do_load_flat_binary(struct linux_binprm * bprm, struct pt_regs * regs)
{
	struct {
		unsigned long u0;
		unsigned long u1;
		unsigned long magic;
		unsigned long mapsize;
	} *ex;
	unsigned long p = bprm->p;
	unsigned long raddr;
	struct file * file;
	int fd;

	(void*)ex = (void*)bprm->buf;
	if (ex->magic != COM_MAGIC)
		return -ENOEXEC;

	if (ex->mapsize < PAGE_SIZE)
		ex->mapsize = PAGE_SIZE;

	raddr = current->rlim[RLIMIT_DATA].rlim_cur;
	if ((raddr < RLIM_INFINITY) && (ex->mapsize > raddr))
		return -ENOMEM;

	fd = flush_old_exec(bprm);
	if (fd)
		return fd;

	current->personality = PER_LINUX;

	current->mm->start_code =
	current->mm->start_data =
		COM_START;
	current->mm->end_code =
	current->mm->end_data =
	current->mm->start_brk =
	current->mm->brk =
		COM_START + ex->mapsize;

	current->mm->rss = 0;
	current->mm->mmap = NULL;

	compute_creds(bprm);
	current->flags &= ~PF_FORKNOEXEC;

	raddr = -EACCES;
	fd = open_dentry(bprm->dentry, O_RDONLY);
	if (fd < 0)
		return fd;
	file = fget(fd);
	if (!file->f_op || !file->f_op->mmap)
	{
bfe:
		fput(file);
		sys_close(fd);
		send_sig(SIGKILL, current, 0);
		return raddr;
	}
	raddr = do_mmap(file, COM_START, PAGE_ALIGN(ex->mapsize),
		PROT_READ | PROT_EXEC,
		MAP_FIXED | MAP_PRIVATE | MAP_DENYWRITE | MAP_EXECUTABLE, 0);
	if (raddr != COM_START)
		goto bfe;

	fput(file);
	sys_close(fd);

	if (current->exec_domain && current->exec_domain->module)
		__MOD_DEC_USE_COUNT(current->exec_domain->module);
	if (current->binfmt && current->binfmt->module)
		__MOD_DEC_USE_COUNT(current->binfmt->module);

	current->exec_domain = lookup_exec_domain(current->personality);
	current->binfmt = &flat_format;

	if (current->exec_domain && current->exec_domain->module)
		__MOD_INC_USE_COUNT(current->exec_domain->module);
	if (current->binfmt && current->binfmt->module)
		__MOD_INC_USE_COUNT(current->binfmt->module);

	p = setup_arg_pages(p, bprm);
	p = (unsigned long) create_flat_tables((char *)p, bprm);

	current->mm->start_stack = p;
	start_thread(regs, (ex->u0) ? COM_START : ex->u1, p);

	regs->edi = regs->esi = regs->esp;
	regs->ebx = 1;
	regs->ecx = 0;
	regs->edx = COM_START;

	if (current->flags & PF_PTRACED)
		send_sig(SIGTRAP, current, 0);

	return 0;
}

static int load_flat_binary(struct linux_binprm * bprm, struct pt_regs * regs)
{
	int retval;

	MOD_INC_USE_COUNT;
	retval = do_load_flat_binary(bprm, regs);
	MOD_DEC_USE_COUNT;
	return retval;
}

int init_module(void)
{
	return register_binfmt(&flat_format);
}

void cleanup_module( void)
{
	unregister_binfmt(&flat_format);
}
